# frozen_string_literal: true

require "yieldable/version"

# = Yieldable
#
# This is meta-mixin, which makes your module, class or it's instance to respond
# to method +to_proc+, for example, to easily build instances during iteration:
#
#   ["1", "2", "3"].map(&SomeObject)
#
# It is 2 different ways to use this module: +extend+ or +prepend+ it to your class.
#
# == <tt>extend Yieldable</tt>
#
#   class SomeObject
#     extend Yieldable # => singleton_class.to_proc = SomeObject.method(:call).to_proc
#   end
#
# This way defines an attribute reader +to_proc+ in a singleton class.
# It sets a proc only once for a class, so it will not generate a new +Proc+ object
# every time, like if you would use just a <tt>method(:call).to_proc</tt>.
# By default it uses the +call+ class method as a source method for a proc. You
# can change the source method by using brackets:
#
#   class SomeObject
#     extend Yieldable[:new] # => singleton_class.to_proc = SomeObject.method(:new).to_proc
#   end
#
# It's better to ensure that a source method is defined _before_ extending a class with +Yieldable+.
# In the other way it will add the +singletone_method_added+ hook and wait. After a source method
# will be added, the hook method will be removed.
#
# You can avoid to removing the hook with an optional <tt>once: false</tt> passed into the brackets:
#
#   extend Yieldable[:new, once: false]
#
# == <tt>prepend Yieldable</tt>
#
# This way defines a +to_proc+ instance method once for a class, but a proc will be generated once for
# each object. You can also use a brackets like within an +extend+ way.
module Yieldable
  # Creates configured mixin.
  #
  # @param method_name [Symbol] A source method's name
  # @param once [Boolean] Will be a proc defined only once? (by default: +true+)
  # @return [Module] the configured mixin
  def self.[](method_name = :call, once: true)
    Mixin.new(method_name, once: once)
  end

  # Utility to resolve method.
  def self.resolve_class_method(base, method_name)
    if base.respond_to?(method_name)
      base.method(method_name)
    else
      base.singleton_method(method_name)
    end
  end

  # Utility to set proc for the mixn.
  def self.set_proc(base, mixin)
    mixin.proc = resolve_class_method(base, mixin.method_name).to_proc
    base.instance_variable_set(:@to_proc, mixin.proc)
    base
  end

  module ToProcReader
    attr_reader :to_proc
  end

  module WaitMethodDefinition
    def singleton_method_added(id)
      if id == YieldableMixin.method_name
        begin
          Yieldable.set_proc(self, YieldableMixin)
        ensure
          singleton_class.remove_method(:singleton_method_added) if YieldableMixin.once
        end
      end

      super(id)
    end
  end

  module Inheritance
    def self.extended(base)
      Yieldable.set_proc(base, base.const_get(:YieldableMixin))
    end

    # Overrides inheritance to set separate YieldableMixin module for a children class.
    def inherited(subclass)
      super
      subclass.const_set(:YieldableMixin, subclass.superclass.const_get(:YieldableMixin).dup)
      Yieldable.set_proc(subclass, subclass.const_get(:YieldableMixin))
    end
  end

  module InstanceMethodsMixin
    def to_proc
      @to_proc ||= method(self.class.const_get(:YieldableMixin).method_name).to_proc
    end
  end

  class Mixin < Module
    attr_reader :method_name, :once
    attr_accessor :proc

    def initialize(method_name = :call, once: true)
      @method_name = method_name
      @once = once
      @proc = nil
      super()
    end

    ##
    # Hook to invoke on `extend Yieldable`.
    def extend_object(base)
      return unless write_once_in_constant(base)
      base.extend(ToProcReader)         unless base.singleton_class.method_defined?(:to_proc)
      base.extend(Inheritance)          if base.singleton_class.method_defined?(@method_name)
      base.extend(WaitMethodDefinition) if wait_method_definition?
      super
    end

    def prepend_features(base)
      return unless write_once_in_constant(base)
      base.prepend InstanceMethodsMixin
      super
    end

    def append_features(base)
      return unless write_once_in_constant(base)
      base.include InstanceMethodsMixin
      super
    end

    private

    def write_once_in_constant(base)
      base.const_set(:YieldableMixin, self) unless base.const_defined?(:YieldableMixin)
    end

    def wait_method_definition?
      @proc.nil? || !@once
    end
  end

  CallMixin = Mixin.new(:call)

  class << self
    private

    def extend_object(base)
      CallMixin.extend_object(base)
    end

    def prepend_features(base)
      CallMixin.prepend_features(base)
    end

    def append_features(base)
      CallMixin.append_features(base)
    end
  end
end